pub use conducted_benchmark::ConductedBenchmark;
pub use file_to_benchmark::FileToBenchmark;

use prettytable::Table;

mod conducted_benchmark;
mod file_to_benchmark;
mod row_builder;
mod row_building;

use crate::{
    cli::BenchmarkCli,
    solving_given::{solve_given, NoSolutionFound},
    AppError, AppResult,
};
use anyhow::anyhow;
use std::time::{Duration, Instant};

pub type BenchmarkResult = Result<ConductedBenchmark, NoSolutionFound>;

pub fn execute_benchmark(args: &BenchmarkCli) -> AppResult<Table> {
    let loaded = load_benchmarks(args)?;

    let benchmarked = loaded.into_iter().map(solve_and_keep_track_of_runtime);

    let mut sum: Duration = Duration::ZERO;
    let mut count: u64 = 0;
    let after_header = row_building::create_rows_for_every_solutions(benchmarked, |to_add| {
        sum += to_add;
        count += 1;
    });
    let average = calc_average(sum, count);

    let mut table = Table::new();
    let rows = after_header
        .into_iter()
        .chain(row_building::create_sum_row(sum))
        .chain(row_building::create_average_row(average))
        .chain(row_building::create_header());
    for next_row in rows {
        table.add_row(next_row);
    }

    Ok(table)
}

fn solve_and_keep_track_of_runtime(
    (benchmark, content): (FileToBenchmark, String),
) -> Result<ConductedBenchmark, NoSolutionFound> {
    let (day, task) = (benchmark.given_day(), benchmark.given_task());
    let before = Instant::now();
    let actual_ouput = solve_given(day, task, &content)?;
    let after = Instant::now();

    let how_long = after - before;

    let result = ConductedBenchmark::new(benchmark, actual_ouput, how_long);
    Ok(result)
}

fn load_benchmarks(args: &BenchmarkCli) -> AppResult<Vec<(FileToBenchmark, String)>> {
    args.files()
        .iter()
        .flat_map(|path| {
            let content = std::fs::read_to_string(path).map_err(|error| {
                anyhow!(
                    "Could not read benchmark file at {:?}\n\
                    Details: {}",
                    path,
                    error
                )
            })?;
            let parsed: Vec<FileToBenchmark> = ron::from_str(&content).map_err(|error| {
                anyhow!(
                    "Content of benchmark file at ({:?}) is in an invalid ron format\n\
                Details of invalid format: {}",
                    path,
                    error
                )
            })?;
            Ok::<_, AppError>(parsed)
        })
        .flatten()
        .map(|benchmark| {
            let where_to_look = benchmark.where_to_look();
            let content = std::fs::read_to_string(where_to_look).map_err(|error| {
                anyhow!(
                    "Could not read puzzle input file at {:?} for day {} and task {}\n\
                    Details: {}",
                    where_to_look,
                    benchmark.given_day(),
                    benchmark.given_task(),
                    error
                )
            })?;
            Ok((benchmark, content))
        })
        .collect()
}

fn calc_average(sum: Duration, nanos_count: u64) -> Duration {
    let average_raw = sum.div_duration_f64(Duration::from_nanos(nanos_count));
    Duration::from_nanos(average_raw.round() as u64)
}
