use std::{borrow::Cow, path::PathBuf, time::Duration};

const AVERAGE: &str = "Average";
const TOTAL: &str = "total";

use prettytable::{color, Row};

use crate::{
    benchmarking::ConductedBenchmark,
    cli::{GivenDay, GivenTask},
    constants,
    solving_given::NoSolutionFound,
};

type OptionalContent = Option<Cow<'static, str>>;

#[derive(Debug, Default)]
pub struct RowBuilder {
    day: OptionalContent,
    task: OptionalContent,
    taken_time: OptionalContent,
    actual_result: OptionalContent,
    expected_result: OptionalContent,
    path_to_input: OptionalContent,
}

impl From<NoSolutionFound> for RowBuilder {
    fn from(value: NoSolutionFound) -> Self {
        match value {
            NoSolutionFound::DayNotFound(day) => Self::no_day_found(day),
            NoSolutionFound::TaskNotFound { day, task } => Self::no_task_found(day, task),
        }
    }
}

impl From<ConductedBenchmark> for RowBuilder {
    fn from(value: ConductedBenchmark) -> Self {
        Self::new(
            value.day(),
            value.task(),
            value.how_long(),
            value
                .maybe_different_actual_output()
                .map(ToString::to_string)
                .unwrap_or_default(),
            value.expected_output().to_string(),
            value.path_of_input().to_path_buf(),
        )
    }
}
fn into_borrowed_cell(some_thing: &'static str) -> OptionalContent {
    Some(Cow::Borrowed(some_thing))
}

fn into_owned_cell(some_thing: impl ToString) -> OptionalContent {
    Some(Cow::Owned(some_thing.to_string()))
}
fn into_owned_cell_with<T>(some_thing: T, on_convert: impl Fn(T) -> String) -> OptionalContent {
    Some(Cow::Owned(on_convert(some_thing)))
}
impl RowBuilder {
    pub fn new(
        day: GivenDay,
        task: GivenTask,
        taken_time: Duration,
        actual_result: String,
        expected_result: String,
        path: PathBuf,
    ) -> Self {
        Self {
            day: into_owned_cell(day),
            task: into_owned_cell(task),
            taken_time: into_owned_cell_with(
                taken_time,
                ConductedBenchmark::convert_duration_to_secs_and_mili_txt,
            ),
            actual_result: into_owned_cell(actual_result),
            expected_result: into_owned_cell(expected_result),
            path_to_input: into_owned_cell_with(path, |path| path.to_string_lossy().to_string()),
        }
    }

    pub fn no_day_found(wrong_day: GivenDay) -> Self {
        let day = into_owned_cell_with(wrong_day, |to_convert| {
            format!("Solution found for day {}", to_convert)
        });
        Self {
            day,
            ..Default::default()
        }
    }

    pub fn no_task_found(right_day: GivenDay, wrong_task: GivenTask) -> Self {
        let day = into_owned_cell(right_day);
        let task = into_owned_cell_with(wrong_task, |to_convert| {
            format!("No solution found for task {}", to_convert)
        });
        Self {
            day,
            task,
            ..Default::default()
        }
    }

    pub fn header() -> Self {
        Self {
            day: into_borrowed_cell("Day"),
            task: into_borrowed_cell("Task"),
            taken_time: into_borrowed_cell("How long (Seconds:Nano)"),
            actual_result: into_borrowed_cell("Actual"),
            expected_result: into_borrowed_cell("Expected"),
            path_to_input: into_borrowed_cell("Used input file"),
        }
    }

    pub fn average(average: Duration) -> Self {
        Self::new_aggreate(AVERAGE, average)
    }

    pub fn total(total: Duration) -> Self {
        Self::new_aggreate(TOTAL, total)
    }

    fn this_or_placeholder(value: OptionalContent) -> prettytable::Cell {
        let value = value.unwrap_or(Cow::Borrowed(constants::PLACEHOLDER_IN_BENCHMARK_REPORTS));
        prettytable::Cell::new(&value)
    }
    fn color_red_if_non_static_value(actual_result: OptionalContent) -> prettytable::Cell {
        let has_different_acutual_than_expected_and_is_no_label = actual_result
            .as_ref()
            .is_some_and(|might_be_borrow| matches!(might_be_borrow, Cow::Owned(_)));
        let tmp = Self::this_or_placeholder(actual_result);
        if has_different_acutual_than_expected_and_is_no_label {
            tmp.with_style(prettytable::Attr::ForegroundColor(color::RED))
        } else {
            tmp
        }
    }
    pub fn into_row(self) -> Row {
        let might_colored_red_actual = Self::color_red_if_non_static_value(self.actual_result);

        let (day, task, taken_time, actual_result, expected_result, path) = (
            Self::this_or_placeholder(self.day),
            Self::this_or_placeholder(self.task),
            Self::this_or_placeholder(self.taken_time),
            might_colored_red_actual,
            Self::this_or_placeholder(self.expected_result),
            Self::this_or_placeholder(self.path_to_input),
        );

        prettytable::Row::new(vec![
            day,
            task,
            taken_time,
            actual_result,
            expected_result,
            path,
        ])
    }

    fn new_aggreate(label: &str, aggregate: Duration) -> Self {
        let (day, task) = (into_owned_cell(label), into_owned_cell(label));
        let taken_time = into_owned_cell_with(
            aggregate,
            ConductedBenchmark::convert_duration_to_secs_and_mili_txt,
        );
        Self {
            day,
            task,
            taken_time,
            ..Default::default()
        }
    }
}
