use std::{cmp::Reverse, collections::HashMap};

use non_zero_count::NonZeroCount;
mod non_zero_count;

#[derive(Debug, PartialEq, Eq, PartialOrd, Ord, Clone, Copy, Default)]
pub enum HandKind {
    /// High card, where all cards' labels are distinct: 23456
    #[default]
    HighCard,
    /// One pair, where two cards share one label, and the other three cards have a different label from the pair and each other: A23A4
    OnePair,
    /// Two pair, where two cards share one label, two other cards share a second label, and the remaining card has a third label: 23432
    TwoPair,
    /// Three of a kind, where three cards have the same label, and the remaining two cards are each different from any other card in the hand: TTT98
    Three,
    /// Full house, where three cards have the same label, and the remaining two cards share a different label: 23332
    FullHouse,
    /// Four of a kind, where four cards have the same label and one card has a different label: AA8AA
    Four,
    /// Five of a kind, where all five cards have the same label: AAAAA
    Five,
}

impl From<(&'_ str, char)> for HandKind {
    fn from((characters, joker_char): (&'_ str, char)) -> Self {
        fn get_joker_count(joker: char, symbols: &str) -> Option<NonZeroCount> {
            let count = symbols.chars().filter(|&next| next == joker).count();
            NonZeroCount::new(count)
        }
        fn use_jokers(
            kind: HandKind,
            how_many_jokers: NonZeroCount,
        ) -> (HandKind, Option<NonZeroCount>) {
            match kind {
                HandKind::HighCard => (HandKind::OnePair, how_many_jokers.decrement()),
                HandKind::OnePair => (HandKind::Three, how_many_jokers.decrement()),
                HandKind::TwoPair => (HandKind::FullHouse, how_many_jokers.decrement()),
                HandKind::Three => (HandKind::Four, how_many_jokers.decrement()),
                HandKind::FullHouse => (HandKind::Four, how_many_jokers.decrement()),
                HandKind::Four | HandKind::Five => (HandKind::Five, Some(how_many_jokers)),
            }
        }

        let duplicated_counted = calc_duplicate_counted(characters, |next| next != joker_char);

        let mut current_kind = choose_kind_from(&duplicated_counted);
        let mut joker_count = get_joker_count(joker_char, characters);

        while joker_count.is_some() && current_kind != HandKind::Five {
            let (new_kind, new_joker_count) = use_jokers(
                current_kind,
                joker_count.expect("While loop only iterates if there is some joke count left"),
            );
            current_kind = new_kind;
            joker_count = new_joker_count;
        }
        current_kind
    }
}

impl From<&'_ str> for HandKind {
    fn from(value: &'_ str) -> Self {
        let duplicated_counted = calc_duplicate_counted(value, |_| true);
        choose_kind_from(&duplicated_counted)
    }
}

fn calc_duplicate_counted(
    characters: &str,
    on_consider_next_char: impl Fn(char) -> bool,
) -> Vec<usize> {
    const FIRST_ENCOUNTERED: usize = 1;

    let mut count_of_same: HashMap<char, usize> = HashMap::with_capacity(10);

    for next_char in characters.chars() {
        if on_consider_next_char(next_char) {
            count_of_same
                .entry(next_char)
                .and_modify(|count| *count += 1)
                .or_insert(FIRST_ENCOUNTERED);
        }
    }

    {
        let mut to_sort: Vec<usize> = count_of_same.into_values().collect();
        to_sort.sort_by_key(|&count| Reverse(count));
        to_sort
    }
}

fn choose_kind_from(to_choose_from: &[usize]) -> HandKind {
    fn calc_kind(values: &[usize]) -> Option<HandKind> {
        let max = values.first()?;

        let kind = match max {
            5 => HandKind::Five,
            4 => HandKind::Four,
            under_five_or_four => {
                let opt_second_after_max = values.get(1);
                match opt_second_after_max {
                    Some(second_after_max) => match (under_five_or_four, second_after_max) {
                        (3, 2) => HandKind::FullHouse,
                        (3, _) => HandKind::Three,
                        (2, 2) => HandKind::TwoPair,
                        (2, 1) => HandKind::OnePair,
                        _no_duplicates => HandKind::HighCard,
                    },
                    None => match under_five_or_four {
                        3 => HandKind::Three,
                        2 => HandKind::OnePair,
                        _no_duplicates => HandKind::HighCard,
                    },
                }
            }
        };

        Some(kind)
    }

    calc_kind(to_choose_from).unwrap_or_else(|| {
        
        HandKind::default()
    })
}

#[cfg(test)]
mod testing {
    use crate::day7::{
        categorized_hand::CategorizedHand, dealt_hand::DealtHand, hand_kind::HandKind,
    };

    #[test]
    fn from_dealt_hand_to_categorized_hand() {
        fn assert_case(given: DealtHand, expected: HandKind) {
            let expected_hand = CategorizedHand::new(expected, given.clone());
            let actual = given.into();
            assert_eq!(expected_hand, actual);
        }
        assert_case(DealtHand::new("AAAAA", 20), HandKind::Five);
        assert_case(DealtHand::new("AA8AA", 20), HandKind::Four);
        assert_case(DealtHand::new("23332", 20), HandKind::FullHouse);
        assert_case(DealtHand::new("TTT98", 20), HandKind::Three);
        assert_case(DealtHand::new("23456", 20), HandKind::HighCard);
        assert_case(DealtHand::new("23432", 20), HandKind::TwoPair);
        assert_case(DealtHand::new("A23A4", 20), HandKind::OnePair);
    }

    #[test]
    fn from_cards_to_kind_with_joker() {
        const JOKER: char = 'J';
        fn assert_case(given: &str, expected: HandKind) {
            let actual: HandKind = (given, JOKER).into();
            assert_eq!(expected, actual, "Given: {:?}", given);
        }

        //32T3K is still the only one pair;
        //it doesn't contain any jokers, so its strength doesn't increase.
        assert_case("32T3K", HandKind::OnePair);
        //KK677 is now the only two pair, making it the second-weakest hand.
        assert_case("KK677", HandKind::TwoPair);
        //T55J5, KTJJT, and QQQJA are now all four of a kind! T55J5 gets rank 3,
        //QQQJA gets rank 4, and KTJJT gets rank 5.
        assert_case("T55J5", HandKind::Four);
        assert_case("KTJJT", HandKind::Four);
        assert_case("QQQJA", HandKind::Four);

        // from highcard
        assert_case("1234J", HandKind::OnePair);
        assert_case("123JJ", HandKind::Three);
        assert_case("12JJJ", HandKind::Four);
        assert_case("1JJJJ", HandKind::Five);
        assert_case("JJJJJ", HandKind::Five);

        // from four
        assert_case("QQQQJ", HandKind::Five);
        assert_case("QQQQQ", HandKind::Five);

        // from three
        assert_case("QQQJJ", HandKind::Five);

        // From two pairs
        assert_case("QQTTJ", HandKind::FullHouse);
    }
}
